"""Implementation of wet-paper codes.

Author: Martin Benes
Affiliation: University of Innsbruck
"""

import numpy as np


def soliton(
    m: int = 100,
    *,
    robust: bool = True,
    c: float = .1,
    delta: float = .5,
) -> np.ndarray:
    """Calculates probabilities of Soliton distribution.

    Taken from Jessica's book, Eqs. 9.5 and 9.6.

    :param m: number of rows/message bits
    :type m: int
    :param robust: Set False to use ideal Soliton. By default True (robust).
    :type robust: bool
    :param c: constant parameter, 0.1 by default
    :type c: float
    :param delta: failure probability, 0.5 by default
    :type delta: float
    :return: probabilities corresponding to 1:m
    :rtype: `np.ndarray <https://numpy.org/doc/stable/reference/generated/numpy.ndarray.html>`__

    :Example:

    >>> i = np.arange(1, 101)
    >>> p_i = soliton(len(i))
    """
    # Soliton distribution
    i = np.arange(1, m+1)
    nu = np.concatenate([[1/m], 1/(i[1:]*i[:-1])])

    # robust Soliton distribution
    if robust:
        T = c * np.log(m / delta) * np.sqrt(m)
        mT = int(np.floor(m / T))
        # print(m, mT)
        assert m > mT
        tau = np.concatenate([
            T / (i[:mT-1] * m),
            [T * np.log(T / delta) / m],
            np.zeros(m - mT),
        ])
        p = nu + tau
    else:
        p = nu

    #
    return p / np.sum(p)


def generate_H(
    m: int,
    n: int = None,
    *,
    c: float = .1,
    delta: float = .5,
    seed: int = None,
) -> np.ndarray:
    """Generates parity-check matrix according to robust soliton distribution.

    :param m: number of rows/message bits
    :type m: int
    :param n: number of columns/cover elements
    :type n: int
    :param c: constant parameter, 0.1 by default
    :type c: float
    :param delta: failure probability, 0.5 by default
    :type delta: float
    :return: generated parity-check matrix
    :rtype: `np.ndarray <https://numpy.org/doc/stable/reference/generated/numpy.ndarray.html>`__

    :Example:

    >>> H = generate_H(50, delta=.5, c=.1, seed=12345)
    """
    if n is None:
        n = 2**m-1
    # Get the robust Soliton distribution
    rsd = soliton(m, c=c, delta=delta, robust=True)

    # Sample column weights w[1], ..., w[n] from the RSD
    rng = np.random.default_rng(seed)
    column_weights = rng.choice(m, size=n, p=rsd) + 1

    # Generate columns of H
    H = np.zeros((m, n), dtype=int)
    for j in range(n):
        # Create a column with column_weights[j] ones
        ones_positions = rng.choice(m, size=column_weights[j], replace=False)
        H[ones_positions, j] = 1

    return H



"""

Author: Martin Benes, Benedikt Lorch
Affiliation: University of Innsbruck
"""

#
from ._simulate import simulate_single_channel
from ._costmap import compute_cost_adjusted, probability

#
from . import _simulate
from . import _costmap

__all__ = [
    '_simulate',
    '_costmap',
    'simulate_single_channel',
]



"""

Author: Martin Benes, Benedikt Lorch
Affiliation: University of Innsbruck
"""

import numpy as np
import scipy.optimize

EPS = np.finfo(np.float64).eps
"""small numerical constant"""


def _entropy(*ps, base: float = 2) -> float:
    """Computes n-ary entropy.

    :param ps: Probability tensors.
    :type ps:
    :param base:
    :type base:
    :param base: unit of information, bits by default
    :type base: float
    :return: Entropy.
    :rtype: float

    :Example:

    >>> # TODO
    """
    # probabilities
    px = np.array(list(ps))
    assert np.all((-EPS <= px) & (px <= 1+EPS)), 'invalid probabilities'
    assert np.isclose(np.sum(px, axis=0), 1).all(), 'denormalized probabilities'
    # entropy
    px[px <= 0] = 1  # avoid log(0)
    if base == 2:  # makes default faster
        H = -(px * np.log2(px))
    else:
        H = -(px * np.log(px) / np.log(base))
    return np.nansum(H)


def entropy(*ps, base: int = 2):
    """Computes entropy with complement probability added.

    :param p: Probability tensor.
    :type p: `np.ndarray <https://numpy.org/doc/stable/reference/generated/numpy.ndarray.html>`__
    :param base: unit of information, bits by default
    :type base: float
    :return: Entropy
    :rtype: float

    :Example:

    >>> # TODO
    """
    px = np.array(list(ps))
    px0 = 1-np.sum(px, axis=0)
    return _entropy(*px, px0, base=base)

# def entropy(p: np.ndarray, p2: np.ndarray = None) -> float:
#     """Computes entropy with complement probability added.

#     :param p: Probability tensor.
#     :type p: `np.ndarray <https://numpy.org/doc/stable/reference/generated/numpy.ndarray.html>`__
#     :param p2: Probability tensor.
#     :type p2: `np.ndarray <https://numpy.org/doc/stable/reference/generated/numpy.ndarray.html>`__
#     :return: Entropy
#     :rtype: float

#     :Example:

#     >>> # TODO
#     """
#     if p2 is None:
#         # Binary entropy
#         return _entropy(p, 1-p)
#     else:
#         # Ternary entropy
#         return _entropy(p, p2, 1-p-p2)


def inv_entropy(
    h: float,
    *,
    xtol: float = 1e-4,
) -> float:
    """Computes inverse entropy.

    :param h:
    :type h:
    :return:
    :rtype: float

    :Example:

    >>> # TODO
    """
    # global EPS

    def mse_H_eq_h(p) -> float:
        """SS error from H(p)=h."""
        return (_entropy(p, 1-p) - h)**2

    return scipy.optimize.fminbound(mse_H_eq_h, 0, .5, xtol=xtol)


def matlab_round(x):
    """
    Helper method to mimic Matlab rounding
    |x| >= 0.5 => round away from zero
    :param x: ndarray
    :return: rounded ndarray
    """

    # Preserve the input sign
    x_sign = np.sign(x)

    # Convert to absolute value
    x_abs = np.abs(x)

    # Round towards zero
    x_abs_floor = np.floor(x_abs)

    # Round away from zero
    x_abs_ceil = np.ceil(x_abs)

    # Compute difference between value and floored value
    abs_floor_diff = x_abs - x_abs_floor

    # Condition for rounding away from zero
    mask_ceil = (abs_floor_diff >= 0.5) | np.isclose(abs_floor_diff, 0.5)

    # Ceil or floor
    x_rounded = np.where(mask_ceil, x_abs_ceil, x_abs_floor)

    # Restore sign
    x_rounded *= x_sign

    return x_rounded



"""

Author: Martin Benes, Benedikt Lorch
Affiliation: University of Innsbruck
"""

# distortion
from . import _costmap
from ._costmap import compute_cost_adjusted

# simulation
from . import _simulate
from ._simulate import simulate_single_channel

__all__ = [
    '_costmap',
    '_simulate',
    'compute_cost_adjusted',
    'simulate_single_channel',
]




import pkg_resources

#
from . import coding
from . import simulate
from . import tools

# spatial
from . import hill
from . import hugo
from . import lsb  # can be also used for JPEG
from . import mipod
from . import suniward
from . import wow
from . import ws

# JPEG
from . import ebs
from . import F5
from . import juniward
from . import nsF5
from . import uerd


# abbreviations of enum
JUNIWARD_ORIGINAL = juniward.Implementation.JUNIWARD_ORIGINAL
JUNIWARD_FIX_OFF_BY_ONE = juniward.Implementation.JUNIWARD_FIX_OFF_BY_ONE
EBS_ORIGINAL = ebs.Implementation.EBS_ORIGINAL
EBS_FIX_WET = ebs.Implementation.EBS_FIX_WET
LSB_REPLACEMENT = lsb.Change.LSB_REPLACEMENT
LSB_MATCHING = lsb.Change.LSB_MATCHING
Location = lsb.Location
LOCATION_PERMUTED = lsb.Location.LOCATION_PERMUTED
LOCATION_SEQUENTIAL = lsb.Location.LOCATION_SEQUENTIAL
LOCATION_SELECTED = lsb.Location.LOCATION_SELECTED
MiPOD_ORIGINAL = mipod.Implementation.MiPOD_ORIGINAL
MiPOD_FIX_WET = mipod.Implementation.MiPOD_FIX_WET
DISTORTION_LIMITED_SENDER = simulate.DISTORTION_LIMITED_SENDER
PAYLOAD_LIMITED_SENDER = simulate.PAYLOAD_LIMITED_SENDER
PLS = PAYLOAD_LIMITED_SENDER
DLS = DISTORTION_LIMITED_SENDER
ATTACKER_INDIFFERENT = tools.ATTACKER_INDIFFERENT
ATTACKER_OMNISCIENT = tools.ATTACKER_OMNISCIENT

# package version
try:
    __version__ = pkg_resources.get_distribution("conseal").version
except pkg_resources.DistributionNotFound:
    __version__ = None

__all__ = [
    'F5',
    'hill',
    'hugo',
    'juniward',
    'lsb',
    'mipod',
    'nsF5',
    'uerd',
    'wow',
    'simulate',
    'tools',
    'JUNIWARD_ORIGINAL',
    'JUNIWARD_FIX_OFF_BY_ONE',
    'EBS_ORIGINAL',
    'EBS_FIX_WET',
    'LSB_REPLACEMENT',
    'LSB_MATCHING',
    '__version__',
]



"""
High-level interface of simulated S-UNIWARD embedding.


Author: Martin Benes, Benedikt Lorch
Affiliation: University of Innsbruck
"""

import numpy as np

from . import _costmap
from ..simulate import _ternary


def simulate_single_channel(
    x0: np.ndarray,
    alpha: float,
    *,
    sigma: float = 1,
    wet_cost: float = 10 ** 8,
    **kw,
) -> np.ndarray:
    """Simulates S-UNIWARD embedding into a single channel.

    :param x0: uncompressed (pixel) cover image,
        of shape [height, width]
    :type x0: `np.ndarray <https://numpy.org/doc/stable/reference/generated/numpy.ndarray.html>`__
    :param alpha: embedding rate,
        in bits per pixel
    :type alpha: float
    :param sigma: stabilizing constant. Small sigmas make the embedding very sensitive to the image content. Large sigmas smooth out the embedding change probabilities.
    :type sigma: float
    :param wet_cost: cost for unembeddable coefficients
    :type wet_cost: float
    :return: stego pixels,
        of shape [height, width, channels]
    :rtype: `np.ndarray <https://numpy.org/doc/stable/reference/generated/numpy.ndarray.html>`__

    :Example:

    >>> x1 = cl.suniward.simulate_single_channel(
    ...   x0=x0,  # pixels
    ...   alpha=0.4,  # embedding rate
    ...   seed=12345)  # seed
    """
    if alpha == 0:
        return x0

    # compute distortion
    rhos = _costmap.compute_cost_adjusted(
        x0=x0,
        sigma=sigma,
        wet_cost=wet_cost
    )

    # simulator
    delta = _ternary.ternary(
        rhos=rhos,
        alpha=alpha,
        n=np.prod(x0.shape),
        **kw,
    )
    return x0 + delta.astype('uint8')




"""

Author: Martin Benes, Benedikt Lorch
Affiliation: University of Innsbruck
"""

import numpy as np
from typing import Tuple

from . import _binary
from . import _ternary
from ._binary import binary
from ._ternary import ternary

from ._optim import get_p, average_payload, average_distortion, Sender

PAYLOAD_LIMITED_SENDER = Sender.PAYLOAD_LIMITED_SENDER
DISTORTION_LIMITED_SENDER = Sender.DISTORTION_LIMITED_SENDER
PLS = PAYLOAD_LIMITED_SENDER
DLS = DISTORTION_LIMITED_SENDER


def simulate(
    rhos: Tuple[np.ndarray],
    alpha: float,
    n: int,
    seed: int = None,
    q: int = None,
    **kw,
) -> Tuple[np.ndarray]:
    """

    :param rhos: either
        a distortion tensor for +-1 change, or
        a tuple with tensors for +1 and -1 change
    :type rho: `np.ndarray <https://numpy.org/doc/stable/reference/generated/numpy.ndarray.html>`__
    :param alpha: embedding rate
    :type alpha: float
    :param n: Cover size.
    :type n: int
    :param seed: random seed for embedding simulator
    :type seed: int
    :return:
    :rtype: `np.ndarray <https://numpy.org/doc/stable/reference/generated/numpy.ndarray.html>`__

    :Example:

    >>> im_dct.Y += cl.simulate.ternary(
    ...     rhos=rhos,        # costs of change
    ...     alpha=0.4,        # alpha
    ...     n=im_dct.Y.size,  # cover size
    ...     seed=12345)       # seed
    """
    # derive q if not given
    if q is None:
        q = len(rhos) + 1

    # ternary
    if q == 3:
        return _ternary.ternary(
            rhos=rhos,
            alpha=alpha,
            n=n,
            seed=seed,
            **kw,
        )
    elif q == 2:
        return _binary.binary(
            rhos=rhos,
            alpha=alpha,
            n=n,
            seed=seed,
            **kw,
        )
    # other
    else:
        raise NotImplementedError(f'{q=} not implemented')


__all__ = [
    '_optim',
    '_ternary',
    'ternary',
    '_binary',
    'binary',
    'get_p',
    'simulate',
    'average_payload',
]



"""

Author: Martin Benes, Benedikt Lorch
Affiliation: University of Innsbruck
"""

import numpy as np
import typing

from .common import entropy


# def embeddable_pixels(
#     x: np.ndarray = None,
#     wet: typing.Tuple = tuple()
# ) -> int:
#     """Get number of embeddable pixels."""
#     # wet elements
#     excluded = sum([(x == w).sum() for w in tuple(wet)])
#     # embeddable elements
#     embeddable = np.prod(x.shape) - excluded
#     return embeddable


# def change_rate(
#     x_c: np.ndarray,
#     x_s: np.ndarray,
#     q: int = 3
# ) -> float:
#     """Estimate change rate in spatial domain"""
#     # no changes
#     if (x_c == x_s).all():
#         return 0.
#     # embeddable elements
#     embeddable = embeddable_pixels(x_c)
#     # binary bound
#     if q == 2:
#         # number of changes
#         changes = (x_c != x_s).sum()
#         # change rate
#         betas = [changes/embeddable]
#     # ternary bound
#     elif q == 3:
#         # number o
#         # number of changes
#         changesP1 = (x_c < x_s).sum()
#         changesM1 = (x_c > x_s).sum()
#         # change rate
#         betas = [
#             changesP1/embeddable,
#             changesM1/embeddable,
#         ]

#     else:
#         raise NotImplementedError(f'{q}-ary code not implemented')

#     return betas


# def embedding_rate(
#     x_c: np.ndarray,
#     x_s: np.ndarray,
#     q: int = 3
# ) -> float:
#     """Estimate embedding rate in spatial domain"""
#     # no changes
#     if (x_c == x_s).all():
#         return 0.
#     # change rates
#     betas = change_rate(x_c, x_s, q=q)
#     # embedding rate
#     alpha = entropy(*betas)
#     return alpha


def daubechies8(
) -> typing.Tuple[typing.Tuple[np.ndarray, np.ndarray], np.ndarray]:
    """Returns Daubechies filter bank."""
    # 1D wavelet filters (Daubechies 8)
    hpdf = np.array([
        -0.0544158422, +0.3128715909, -0.6756307363, +0.5853546837,
        +0.0158291053, -0.2840155430, -0.0004724846, +0.1287474266,
        +0.0173693010, -0.0440882539, -0.0139810279, +0.0087460940,
        +0.0048703530, -0.0003917404, -0.0006754494, -0.0001174768
    ])
    lpdf = np.power(-1, np.arange(len(hpdf)))*np.flip(hpdf)
    # 2D wavelet filters
    F = [
        lpdf.reshape(-1, 1) @ hpdf.reshape(1, -1),
        hpdf.reshape(-1, 1) @ lpdf.reshape(1, -1),
        hpdf.reshape(-1, 1) @ hpdf.reshape(1, -1),
    ]
    return (hpdf, lpdf), F



"""

The equation is from

P. Bas, et al. "Break Our Steganographic System: The Ins and Outs of Organizing BOSS", IH 2011.

Originally described in the STC paper by
T. Filler, et al. "Minimizing Additive Distortion in Steganography Using Syndrome-Trellis Codes", TIFS 2011.


Author: Martin Benes, Benedikt Lorch
Affiliation: University of Innsbruck
"""

import enum
import numpy as np
from typing import Callable, Tuple
import warnings

from .. import tools


class Sender(enum.Enum):
    """Type of sender."""

    PAYLOAD_LIMITED_SENDER = enum.auto()
    """Payload-limited sender."""
    DISTORTION_LIMITED_SENDER = enum.auto()
    """Distortion-limited sender."""


def get_p(
    lbda: float,
    *rhos: np.ndarray,
    add_zero: bool = True,
) -> np.ndarray:
    """Converts distortions into probabilities,
    using Boltzmann-Gibbs distribution

    For more details, see `glossary <https://conseal.readthedocs.io/en/latest/glossary.html#embedding-simulation>`__.

    :param rhos: distortion of embedding choices, e.g. embedding +1 or embedding -1
    :type rhos: tuple
    :param lbda: parameter value
    :type lbda: float
    :param add_zero:
    :type add_zero: bool
    :param p_pm1: probability tensor for changes associated to rhos[0]
        of an arbitrary shape
    :rtype: `np.ndarray <https://numpy.org/doc/stable/reference/generated/numpy.ndarray.html>`__

    :Example:

    >>> # TODO
    """
    # denominator (forced left-associativity)
    denum = 1 if add_zero else 0
    for rho in rhos:
        denum += np.exp(-lbda * rho)
    #
    denum[denum == 0] = tools.EPS
    return np.exp(-lbda * rhos[0]) / denum


def average_payload(
    *,
    ps: Tuple[np.ndarray] = None,
    e: float = None,
    lbda: float = None,
    rhos: Tuple[np.ndarray] = None,
    q: int = None,
) -> Tuple[Tuple[np.ndarray], float]:
    """

    :param ps: Probability maps.
    :type ps: tuple of `np.ndarray <https://numpy.org/doc/stable/reference/generated/numpy.ndarray.html>`__
    :param e: Embedding efficiency. If not provided, optimal coding is assumed.
    :type e: float
    :param lbda: Parameter of the Gibbs distribution, if rhos are given.
    :type lbda: float
    :param rhos: Cost maps. Can be provided instead of `ps`.
    :type rhos: tuple of `np.ndarray <https://numpy.org/doc/stable/reference/generated/numpy.ndarray.html>`__
    :return:
    :rtype: float

    :Example:

    >>> # TODO
    """
    assert (
        (ps is not None and rhos is None or ps is None and rhos is not None)
    ), 'one of ps or rhos must be given'
    assert (
        lbda is not None or ps is not None
    ), 'lbda can be specified only with rhos'
    #
    if ps is None:
        add_zero = True if q is None else len(rhos) == q-1
        ps = [
            get_p(lbda, rhos[i], *rhos[:i], *rhos[i+1:], add_zero=add_zero)
            for i in range(len(rhos))
        ]

    # Imperfect coding - given embedding efficiency
    if e is not None:
        H = np.sum(ps) * e
    # Perfect coding - upper bound efficiency
    elif q is None or len(ps) == q-1:  # no change is zero cost
        H = tools.entropy(*ps)
    else:
        H = tools._entropy(*ps)

    return ps, H


def average_distortion(
    rhos: Tuple[np.ndarray],
    *,
    # e: float = None,
    lbda: float = None,
    ps: Tuple[np.ndarray] = None,
    # q: int = None,
) -> Tuple[Tuple[np.ndarray], float]:
    """

    :param ps: Probability maps.
    :type ps: tuple of `np.ndarray <https://numpy.org/doc/stable/reference/generated/numpy.ndarray.html>`__
    :param e: Embedding efficiency. If not provided, optimal coding is assumed.
    :type e: float
    :param lbda: Parameter of the Gibbs distribution, if rhos are given.
    :type lbda: float
    :param rhos: Cost maps. Can be provided instead of `ps`.
    :type rhos: tuple of `np.ndarray <https://numpy.org/doc/stable/reference/generated/numpy.ndarray.html>`__
    :return:
    :rtype: float

    :Example:

    >>> # TODO
    """
    assert (
        (ps is not None and lbda is None or ps is None and lbda is not None)
    ), 'one of ps or lbda must be given'
    #
    if ps is None:
        # add_zero = True if q is None else len(rhos) == q-1
        add_zero = True
        ps = [
            get_p(lbda, rhos[i], *rhos[:i], *rhos[i+1:], add_zero=add_zero)
            for i in range(len(rhos))
        ]

    # average distortion
    D = np.sum([
        rhos[i] * ps[i]
        for i in range(len(rhos))
    ])
    return ps, D


def get_objective(
    sender: Sender = Sender.PAYLOAD_LIMITED_SENDER,
    e: float = None,
    q: int = None,
) -> Callable:
    def _pls_objective(*args, **kw):
        return average_payload(*args, e=e, q=q, **kw)

    if sender == Sender.PAYLOAD_LIMITED_SENDER:
        # print('selected PLS objective')
        return average_payload if e is None else _pls_objective
    if sender == Sender.DISTORTION_LIMITED_SENDER:
        # print('selected DLS objective')
        assert e is None, 'e not implemented for DLS'
        return average_distortion


def calc_lambda(
    rhos: Tuple[np.ndarray],
    m: int,
    n: int,
    objective: Callable = None,
) -> float:
    """Implements binary search for lambda.

    The i-th element is embedded with a probability of
    p_i = 1/Z exp( -lambda D(X, y_i X_{~i}) ),
    where D is the distortion after modifying the i-th cover element, and Z is a normalization constant.
    This methods determines the lambda to communicate the message of the message length.

    We simulate a payload-limited sender that embeds a fixed average payload while minimizing the average distortion.
    Optimize for the lambda that minimizes the average distortion while transferring the message.

    :param rhos: Tuple of costs.
    :type rho_p1: tuple of `np.ndarray <https://numpy.org/doc/stable/reference/generated/numpy.ndarray.html>`__
    :param m: Message length.
    :type m: int
    :param n: Cover size.
    :type n: int
    :return: Parameter lambda value.
    :rtype: float

    :Example:

    >>> # TODO
    """
    assert n > 0, "Expected cover size greater than 0"

    # Initialize lambda and m3 such that the loop is at least entered once
    # m3 is the total entropy
    l3 = 1000
    m3 = float(m + 1)

    # Initialize iteration counter
    iterations = 0

    # Find the largest l3, s.t., H(m3) <= m
    while m3 > m:

        # Increase l3
        l3 *= 2

        # Compute total entropy m3
        _, m3 = objective(lbda=l3, rhos=rhos)  # objective function

        iterations += 1

        # unbounded = search fails
        if iterations > 15:
            warnings.warn("unbounded distortion, search fails", RuntimeWarning)
            return l3

    # Initialize lower bound to zero
    l1 = 0
    # The lower bound for the message size is n
    m1 = float(n)
    alpha = float(m) / n  # embedding rate

    # Binary search for lambda
    # Relative payload must be within 1e-3 of the required relative payload
    while float(m1 - m3) / n > alpha / 1000 and iterations < 30:
        # Mid of the interval [l1, l3]
        lbda = l1 + (l3 - l1) / 2

        # Calculate entropy at the mid of the interval
        _, m2 = objective(lbda=lbda, rhos=rhos)  # objective function

        # binary search
        if m2 < m:
            # The average payload is too small for the message.
            # We need to decrease the upper bound
            l3 = lbda
            m3 = m2
        else:
            # The average payload exceeds the message length
            # We can increase the lower bound
            l1 = lbda
            m1 = m2

        # Proceed to the next iteration
        iterations = iterations + 1

    if iterations == 30:
        warnings.warn("optimization might not have converged", RuntimeWarning)

    return lbda
